/*
 * i-list-walker -- apply a specified function to all inodes on a filesystem
 *
 * Based on original routines by Larry Philps.  This reimplementation is
 * Copyright 1988 by Rayan Zachariassen, solely to prevent you from selling
 * it or putting your name on it.  Free (gratis) redistribution is encouraged.
 */

#define HAVE_ST_BLKSIZE

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#ifdef __svr4__
#include <sys/vnode.h>
#include <sys/fs/ufs_fs.h>
#include <sys/fs/ufs_inode.h>
#else
# ifdef sgi
#  include <sys/sysmacros.h>
#  include <sys/fs/efs.h>
#  include <sys/syssgi.h>
#  ifndef _STAT_VER
#   undef HAVE_ST_BLKSIZE	/* for IRIX 4.x */
#  endif
#  ifndef IFMT
#   define IFMT S_IFMT		/* for IRIX 4.x */
#  endif
# else
#  include <sys/vnode.h>
#  include <ufs/fs.h>
#  include <ufs/inode.h>
# endif
#endif

extern int	ildebug;
static char	*rawdev();

struct handle {
	int fd;		/* file descriptor for raw dev */
	int (*fn)();	/* user's function */
	dev_t dev;	/* stat info that isn't in the disk inode */
	long blksize;
			/* superblock */
	union {
#ifdef sgi
#define SBLOCK EFS_SUPERBB
#define SBSIZE BBTOB(1)
#define dinode efs_dinode
#define fs_ipg fs_ipcg
		struct efs fs;
#else
		struct fs fs;
#endif
		char dummy[SBSIZE];
	} sb;
};


static int
bio(fn, fd, bno, buf, cnt)
	int (*fn)();	/* read() or write() */
	int fd;
	daddr_t	bno;
	char	*buf;
	int	cnt;
{
#ifdef sgi
	cnt = BTOBB(cnt);
	if (syssgi(fn == write ? SGI_WRITEB : SGI_READB,
	    fd, buf, bno, cnt) != cnt)
		return -1;
#else
	if (lseek(fd, (off_t) bno * DEV_BSIZE, 0) == -1)
		return -1;
	if ((*fn)(fd, buf, cnt) != cnt)
		return -1;
#endif
	return 0;
}

/*
 * i-list-walker
 *
 * Calls the function fn with an inode
 * pointer for every inode on the device.
 * Return value of fn is:
 *	<0: error, halt.
 *	 0: continue
 *	>0: write changes back to disk and continue
 *
 * Return -1 in case of error, +ve if inodes were updated, 0 otherwise.
 */

int
ilw(h, fn)
	struct handle *h;
	int (*fn)();
{
	ino_t	ino;
	int	cg, j, changed, tabsize, status, k;
	daddr_t		iblk;
	struct dinode	*itab;

#ifdef sgi
	tabsize = BBTOB(h->sb.fs.fs_cgisize);
#else
	tabsize = h->sb.fs.fs_ipg * sizeof(struct dinode);
#endif
	if ((itab = (struct dinode *)malloc(tabsize)) == NULL)
		return -1;
	sync();
	status = 0;
	ino = 0;
	for (cg = 0; cg < h->sb.fs.fs_ncg; cg++) {
#ifdef sgi
		iblk = EFS_CGIMIN(&h->sb.fs, cg);
#else
		iblk = fsbtodb(&h->sb.fs, cgimin(&h->sb.fs, cg));
#endif
		if (bio(read, h->fd, iblk, (char *)itab, tabsize)) {
			status = -1;
			break;
		}
		changed = 0;
		for (j = 0; j < h->sb.fs.fs_ipg; j++, ino++) {
			if ((itab[j].di_mode&IFMT) == 0)
				continue;
			k = (*fn)(h, &itab[j], ino);
			if (k < 0) {
				status = -1;
				goto done;
			}
			if (k > 0)
				changed++;
		}
		if (changed) {
			if (bio(write, h->fd, iblk, (char *)itab, tabsize)) {
				status = -1;
				break;
			}
			status++;
		}
	}
    done:
	free(itab);
	return status;
}


/*
 * path is a filesystem mount point.
 * Determine the raw disk device, read the superblock,
 * and return a handle to the result.
 */
void *
ilist_open(path)
	char *path;
{
	char rpath[MAXPATHLEN];
	struct stat st;
	struct handle *h;

	if (ildebug) fprintf(stderr, "ilist_open %s\n", path);
	if (stat(path, &st) < 0)
		return NULL;
	if ((h = (struct handle *)malloc(sizeof(struct handle))) == NULL)
		return NULL;
	h->dev = st.st_dev;
#ifdef HAVE_ST_BLKSIZE
	h->blksize = st.st_blksize;
#else
	h->blksize = 1024;	/* whatever */
#endif
	if (rawdev(path, rpath, &st) == NULL ||
	    (h->fd = open(rpath, 0)) < 0) {
		free(h);
		return NULL;
	}
	sync();
	if (bio(read, h->fd, SBLOCK, (char *)&h->sb.dummy, SBSIZE) ||
#ifdef sgi
	    !IS_EFS_MAGIC(h->sb.fs.fs_magic)) {
#else
            h->sb.fs.fs_magic != FS_MAGIC) {
#endif
		close(h->fd);
		free(h);
		if (ildebug) fprintf(stderr, "bad magic number\n");
		return NULL;
	}
#ifdef sgi
	EFS_SETUP_SUPERB(&h->sb.fs);
#endif
	if (ildebug) fprintf(stderr, "fs contains %d files\n", ilist_nfiles(h));
	return (void *)h;
}

int
ilist_nfiles(h)
	struct handle *h;
{
	return h->sb.fs.fs_ipg * h->sb.fs.fs_ncg;
}

/* Convert a (struct dinode) to a (struct stat) */
static int
inode_to_stat(h, ip, inum)
	struct handle *h;
	struct dinode *ip;
	int inum;
{
	struct stat st;

	/* convert ip to struct stat */
	/* memset(&st, 0, sizeof st); */
        st.st_dev = h->dev;
        st.st_ino = inum;
        st.st_mode = ip->di_mode;
        st.st_nlink = ip->di_nlink;
        st.st_uid = ip->di_uid;
        st.st_gid = ip->di_gid;
        st.st_size = ip->di_size;
        st.st_atime = ip->di_atime;
        st.st_mtime = ip->di_mtime;
        st.st_ctime = ip->di_ctime;
#ifdef HAVE_ST_BLKSIZE
	st.st_blksize = h->blksize;
#endif
#ifdef sgi
# ifdef HAVE_ST_BLKSIZE
	st.st_rdev = ip->di_u.di_dev.ndev;
	/* XXX this is costly and not usually needed */
	/* maybe we should use &~mask */
	st.st_blocks = (st.st_size + h->blksize-1)/h->blksize;
# else
	st.st_rdev = ip->di_u.di_dev;
# endif
#else
# ifdef __svr4__
	st.st_rdev = ip->di_un.di_icom.ic_db[0];	/* gag */
# else
	st.st_rdev = ip->di_rdev;
# endif
	st.st_blocks = ip->di_blocks;
#endif

	return (*h->fn)(h, &st, inum);	/* NB: fn must not return +ve */
}

int
ilist_walk(h, fn)
	struct handle *h;
	int (*fn)();
{
	h->fn = fn;
	return ilw(h, inode_to_stat);
}

void
ilist_close(h)
	struct handle *h;
{
	close(h->fd);
	free(h);
}

/*
 * Find the raw disk device associated with a mount point.
 * This is quite messy.
 */
#ifdef __svr4__
#include <sys/mnttab.h>
#include <sys/mntent.h>
#define mnt_type mnt_fstype
#else
#include <mntent.h>
#endif

static char *
rawdev(path, buf, st)
	char *path, *buf;
	struct stat *st;	/* stat(path) */
{
	char *cp, *dir, *spec, *ret = NULL;
	FILE *mntfp;
	struct stat dst;

#ifdef __svr4__
	struct mnttab mntbuf, *mntp = &mntbuf;
	if ((mntfp = fopen(MNTTAB, "r")) == NULL) return NULL;
	while (getmntent(mntfp, mntp) == 0) {
		dir = mntp->mnt_mountp;
		spec = mntp->mnt_special;
#else
	struct mntent *mntp;
	if ((mntfp = setmntent(MNTTAB, "r")) == NULL) return NULL;
	while ((mntp = getmntent(mntfp)) != NULL) {
		dir = mntp->mnt_dir;
		spec = mntp->mnt_fsname;
#endif
#ifdef MNTOPT_IGNORE
		if (hasmntopt(mntp, MNTOPT_IGNORE))
			continue;
#endif
		if (strcmp(dir, path) != 0) {
#ifdef MNTOPT_DEV
			/*
			 * Avoid needless stat's: if the device numbers
			 * don't match, skip it.
			 */
			if ((cp = hasmntopt(mntp, MNTOPT_DEV)) != NULL &&
			    (cp = strchr(cp, '=')) != NULL) {
				if ((dev_t)strtol(cp+1, NULL, 16) != st->st_dev)
					continue;
			}
#endif
			/*
			 * They still might be the same:
			 * check the dev/ino.  This works
			 * around automounter aliasing problems.
			 */
			if (!(stat(dir, &dst) == 0 &&
			      st->st_dev == dst.st_dev &&
			      st->st_ino == dst.st_ino))
			continue;
		}
#ifdef MNTTYPE_EFS
		if (strcmp(mntp->mnt_type, MNTTYPE_EFS) == 0 &&
		    (cp = hasmntopt(mntp, MNTOPT_RAW)))
                        strcpy(ret = buf, cp + strlen(MNTOPT_RAW)+1);
#else
# ifdef MNTTYPE_UFS
		if (strcmp(mntp->mnt_type, MNTTYPE_UFS) != 0)
			continue;
# endif
		if ((cp = strrchr(spec, '/')) == NULL)
			break; /* Devices must have absolute paths */
		*cp++ = '\0';
		if (strcmp(spec, "/dev/dsk") == 0)
			sprintf(buf, "/dev/rdsk/%s", cp);
		else
			sprintf(buf, "%s/r%s", spec, cp);
		cp[-1] = '/';
		ret = buf;
#endif
		break;
	}
#ifdef __svr4__
	fclose(mntfp);
#else
	endmntent(mntfp);
#endif
	if (ildebug) fprintf(stderr, "rawdev %s\n", ret ? ret : "?");
	return ret;
}
